/* gost.c - GOST public key operations used in public key glue code
 *	Copyright (C) 2019 Paul Wolneykien <manowar@altlinux.org>.
 *
 * This file is part of GnuPG.
 *
 * GnuPG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnuPG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "gpg.h"
#include "../common/util.h"
#include "../common/gost-util.h"
#include "pkglue.h"
#include "main.h"
#include "options.h"

/* Version: 2 for GOST (FIXME, RFC). */
#define GOST_KDF_PARAMS_VERSION 2

typedef enum
  {
    DIGEST_PARAMS_UNSPECIFIED = 0,
    DIGEST_PARAMS_GOSTR3411_94_A = 1, /* variant Verba-O: 1.2.643.2.2.30.1 */
  }
digest_params_t;

typedef enum
  {
    CIPHER_PARAMS_GOST28147_A = 1,  /* variant Verba-O: 1.2.643.2.2.31.1 */
    CIPHER_PARAMS_GOST28147_B = 2,  /* variant 1: 1.2.643.2.2.31.2 */
    CIPHER_PARAMS_GOST28147_C = 3,  /* variant 2: 1.2.643.2.2.31.3 */
    CIPHER_PARAMS_GOST28147_D = 4,  /* variant 3: 1.2.643.2.2.31.4 */
    CIPHER_PARAMS_GOST28147_Z = 5,  /* variant Z: 1.2.643.7.1.2.5.1.1 */
  }
cipher_params_t;

typedef enum
  {
    MAC_PARAMS_UNSPECIFIED = 0,
    MAC_PARAMS_GOST28147_A = 1,  /* variant Verba-O: 1.2.643.2.2.31.1 */
    MAC_PARAMS_GOST28147_B = 2,  /* variant 1: 1.2.643.2.2.31.2 */
    MAC_PARAMS_GOST28147_C = 3,  /* variant 2: 1.2.643.2.2.31.3 */
    MAC_PARAMS_GOST28147_D = 4,  /* variant 3: 1.2.643.2.2.31.4 */
    MAC_PARAMS_GOST28147_Z = 5,  /* variant Z: 1.2.643.7.1.2.5.1.1 */
  }
mac_params_t;

typedef enum
  {
    VKO_7836 = 1,
  }
vko_algo_t;

typedef enum
  {
    KDF_NULL = 0,
    GOST_KDF_CPDIVERS = 1,  /* or KDF_4357 (RFC 4357) */
    GOST_KDF_TREE = 2,      /* or KDF_7836 (RFC 7836) */
  }
kdf_algo_t;

typedef enum
  {
    KEYWRAP_7836,
  }
keywrap_algo_t;

typedef struct
  {
    vko_algo_t vko_algo;   /* VKO algorithm */
    union {
      struct {
        unsigned char ukm_len;  /* UKM length in bytes */
        /* Digest algorithm parameters for VKO */
        digest_algo_t   vko_digest_algo;
        digest_params_t vko_digest_params;
      } vko_7836; /* Parameters of the VKO algorithm in RFC 7836 */
    } vko_params; /* Paramenters of the VKO algorithm */
    kdf_algo_t kdf_algo; /* Key derivation algorithm */
    union {
      struct {
        /* Cipher algorithm parameters for wrapping */
        cipher_algo_t kdf_cipher_algo;
        cipher_params_t kdf_cipher_params;
      } kdf_4357;
      struct {
        unsigned char seed_len; /* Seed length in bytes */
        char *label;      /* KDF tree label */
        unsigned char R;        /* Inerations (1--4) */
        unsigned int L;         /* Generated key length (in bits). */
        /* Digest algorithm parameters for KDF */
        digest_algo_t   kdf_digest_algo;
        digest_params_t kdf_digest_params;
      } kdf_7836; /* Parameters for the KDF tree algorithm in RFC 7836. */
    } kdf_params; /* Paramenters of the KDF algorithm */
    keywrap_algo_t keywrap_algo; /* Key wrapping algorithm */
    union {
      struct {
        /* MAC algorithm parameters for wrapping */
        mac_algo_t   keywrap_mac_algo;
        mac_params_t keywrap_mac_params;
        /* Cipher algorithm parameters for wrapping */
        cipher_algo_t   keywrap_cipher_algo;
        cipher_params_t keywrap_cipher_params;
      } keywrap_7836; /* Key wrapping algorithm parameters in RFC 7836 */
    } keywrap_params; /* Key wrapping algorithm parameters */
  }
gost_kdf_params_t;

/* A table with the default GOST encryption parameters used by
   GnuPG. */
static const struct
{
  const char *oidpfx;  /* Key algorithm OID prefix */
  unsigned int qbits;  /* Key length in bits */
  gost_kdf_params_t params;  /* Corresponding VKO/KDF/KW parameters. */
} gost_kdf_params_table[] =
  /* Note: Must be sorted by ascending values for QBITS.  */
  {
    { "1.2.643.2.2.35.", 256,
      {
        .vko_algo = VKO_7836,
        .vko_params.vko_7836 =
        {
          8, DIGEST_ALGO_GOSTR3411_94, DIGEST_PARAMS_GOSTR3411_94_A
        },
        .kdf_algo = GOST_KDF_CPDIVERS,
        .kdf_params.kdf_4357 =
        {
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_A
        },
        .keywrap_algo = KEYWRAP_7836,
        .keywrap_params.keywrap_7836 =
        {
          MAC_ALGO_GOST28147_IMIT, MAC_PARAMS_GOST28147_A,
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_A
        }
      }
    },
    { "1.2.643.2.2.36.", 256,
      {
        .vko_algo = VKO_7836,
        .vko_params.vko_7836 =
        {
          8, DIGEST_ALGO_GOSTR3411_94, DIGEST_PARAMS_GOSTR3411_94_A
        },
        .kdf_algo = GOST_KDF_CPDIVERS,
        .kdf_params.kdf_4357 =
        {
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_A
        },
        .keywrap_algo = KEYWRAP_7836,
        .keywrap_params.keywrap_7836 =
        {
          MAC_ALGO_GOST28147_IMIT, MAC_PARAMS_GOST28147_A,
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_A
        }
      }
    },
    { "1.2.643.7.1.2.1.1.", 256,
      {
        .vko_algo = VKO_7836,
        .vko_params.vko_7836 =
        {
          8, DIGEST_ALGO_GOSTR3411_12_256, DIGEST_PARAMS_UNSPECIFIED
        },
        .kdf_algo = GOST_KDF_CPDIVERS,
        .kdf_params.kdf_4357 =
        {
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_Z
        },
        .keywrap_algo = KEYWRAP_7836,
        .keywrap_params.keywrap_7836 =
        {
          MAC_ALGO_GOST28147_IMIT, MAC_PARAMS_GOST28147_Z,
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_Z
        }
      }
    },
    { "1.2.643.7.1.2.1.2.", 512,
      {
        .vko_algo = VKO_7836,
        .vko_params.vko_7836 =
        {
          8, DIGEST_ALGO_GOSTR3411_12_256, DIGEST_PARAMS_UNSPECIFIED
        },
        .kdf_algo = GOST_KDF_CPDIVERS,
        .kdf_params.kdf_4357 =
        {
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_Z
        },
        .keywrap_algo = KEYWRAP_7836,
        .keywrap_params.keywrap_7836 =
        {
          MAC_ALGO_GOST28147_IMIT, MAC_PARAMS_GOST28147_Z,
          CIPHER_ALGO_GOST28147, CIPHER_PARAMS_GOST28147_Z
        }
      }
    },
  };

static gpg_error_t
pack_gost_kdf_params (const gost_kdf_params_t *params, unsigned char *buf,
                      size_t *length)
{
  size_t len = 0;

  if (*length < (len + 1)) return GPG_ERR_TOO_SHORT;
  buf[len+0] = params->vko_algo;
  len += 1;

  switch (params->vko_algo)
    {
    case VKO_7836:
      if (*length < (len + 3)) return GPG_ERR_TOO_SHORT;
      buf[len+0] = params->vko_params.vko_7836.ukm_len;
      buf[len+1] = params->vko_params.vko_7836.vko_digest_algo;
      buf[len+2] = params->vko_params.vko_7836.vko_digest_params;
      len += 3;
      break;

    default:
      return GPG_ERR_UNKNOWN_ALGORITHM;
    }

  if (*length < (len + 1)) return GPG_ERR_TOO_SHORT;
  buf[len+0] = params->kdf_algo;
  len += 1;

  switch (params->kdf_algo)
    {
    case GOST_KDF_TREE:
      {
        size_t label_len = strlen (params->kdf_params.kdf_7836.label);
        if (*length < (len + 8 + label_len)) return GPG_ERR_TOO_SHORT;
        buf[len+0] = params->kdf_params.kdf_7836.seed_len;
        /* Bits 15--8, 7--0 of label_len */
        buf[len+1] = (label_len >> 8) & 0xff;
        buf[len+2] = label_len & 0xff;
        memcpy (buf+len+3, params->kdf_params.kdf_7836.label, label_len);
        len += 3 + label_len;
        buf[len+0] = params->kdf_params.kdf_7836.R;
        /* Bits 15--8, 7--0 of L*/
        buf[len+1] = (params->kdf_params.kdf_7836.L >> 8) & 0xff;
        buf[len+2] = params->kdf_params.kdf_7836.L & 0xff;
        buf[len+3] = params->kdf_params.kdf_7836.kdf_digest_algo;
        buf[len+4] = params->kdf_params.kdf_7836.kdf_digest_params;
        len += 5;
      }
      break;

    case GOST_KDF_CPDIVERS:
      buf[len+0] = params->kdf_params.kdf_4357.kdf_cipher_algo;
      buf[len+1] = params->kdf_params.kdf_4357.kdf_cipher_params;
      len += 2;
      break;

    case KDF_NULL:
      break; /* No params */

    default:
      return GPG_ERR_UNKNOWN_ALGORITHM;
    }

  if (*length < (len + 1)) return GPG_ERR_TOO_SHORT;
  buf[len+0] = params->keywrap_algo;
  len += 1;

  switch (params->keywrap_algo)
    {
    case KEYWRAP_7836:
      if (*length < (len + 4)) return GPG_ERR_TOO_SHORT;
      buf[len+0] = params->keywrap_params.keywrap_7836.keywrap_mac_algo;
      buf[len+1] = params->keywrap_params.keywrap_7836.keywrap_mac_params;
      buf[len+2] = params->keywrap_params.keywrap_7836.keywrap_cipher_algo;
      buf[len+3] = params->keywrap_params.keywrap_7836.keywrap_cipher_params;
      len += 4;
      break;

    default:
      return GPG_ERR_UNKNOWN_ALGORITHM;
    }

  *length = len;

  return GPG_ERR_NO_ERROR;
}

void
free_gost_kdf_params (gost_kdf_params_t *params)
{
  if (params)
    {
      switch (params->kdf_algo)
        {
        case GOST_KDF_TREE:
          if (params->kdf_params.kdf_7836.label)
            xfree (params->kdf_params.kdf_7836.label);
          break;
        }
      xfree (params);
    }
}

static gpg_error_t
unpack_gost_kdf_params (byte *packed, gost_kdf_params_t **r_params)
{
  gpg_error_t ret = GPG_ERR_NO_ERROR;
  gost_kdf_params_t *params = NULL;
  unsigned int pos = 0;

  params = xtrymalloc (sizeof (*params));
  if (!params) return gpg_error_from_syserror ();

  memset (params, 0, sizeof (*params));

  params->vko_algo = packed[pos+0];
  pos += 1;

  switch (params->vko_algo)
    {
    case VKO_7836:
      params->vko_params.vko_7836.ukm_len = packed[pos+0];
      params->vko_params.vko_7836.vko_digest_algo = packed[pos+1];
      params->vko_params.vko_7836.vko_digest_params = packed[pos+2];
      pos += 3;
      break;

    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
      goto exit;
    }

  params->kdf_algo = packed[pos+0];
  pos += 1;

  switch (params->kdf_algo)
    {
    case GOST_KDF_TREE:
      {
        params->kdf_params.kdf_7836.seed_len = packed[pos+0];
        /* Bits 15--8, 7--0 of label_len */
        size_t label_len = (packed[pos+1] << 8) | packed[pos+2];
        params->kdf_params.kdf_7836.label = xtrymalloc (label_len + 1);
        if (!params->kdf_params.kdf_7836.label)
          {
            ret = gpg_error_from_syserror ();
            goto exit;
          }
        memcpy (params->kdf_params.kdf_7836.label, packed+pos+3, label_len);
        params->kdf_params.kdf_7836.label[label_len] = '\0';
        pos += 3 + label_len;
        params->kdf_params.kdf_7836.R = packed[pos+0];
        /* Bits 15--8, 7--0 of L*/
        params->kdf_params.kdf_7836.L = (packed[pos+1] << 8) | packed[pos+2];
        params->kdf_params.kdf_7836.kdf_digest_algo = packed[pos+3];
        params->kdf_params.kdf_7836.kdf_digest_params = packed[pos+4];
        pos += 5;
      }
      break;

    case GOST_KDF_CPDIVERS:
      params->kdf_params.kdf_4357.kdf_cipher_algo = packed[pos+0];
      params->kdf_params.kdf_4357.kdf_cipher_params = packed[pos+1];
      pos += 2;
      break;

    case KDF_NULL:
      break; /* No params */

    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
      goto exit;
    }

  packed[pos+0] = params->keywrap_algo;
  pos += 1;

  switch (params->keywrap_algo)
    {
    case KEYWRAP_7836:
      params->keywrap_params.keywrap_7836.keywrap_mac_algo = packed[pos+0];
      params->keywrap_params.keywrap_7836.keywrap_mac_params = packed[pos+1];
      params->keywrap_params.keywrap_7836.keywrap_cipher_algo = packed[pos+2];
      params->keywrap_params.keywrap_7836.keywrap_cipher_params = packed[pos+3];
      pos += 4;
      break;

    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
    }

 exit:
  if (ret != GPG_ERR_NO_ERROR)
    {
      free_gost_kdf_params (params);
      params = NULL;
    }
  else
    *r_params = params;

  return ret;
}

gpg_error_t
pk_gost_get_kdf_params (gcry_mpi_t *pkey, gost_kdf_params_t **r_params)
{
  const unsigned char *kdf_params = NULL;
  size_t kdf_params_size;
  unsigned int nbits;

  kdf_params = gcry_mpi_get_opaque (pkey[2], &nbits);
  kdf_params_size = (nbits+7)/8;

  /* Expect GOST version (see pk_gost_default_params() below) */
  if (kdf_params_size > 2 && kdf_params[1] != GOST_KDF_PARAMS_VERSION)
    return GPG_ERR_BAD_PUBKEY;

  return unpack_gost_kdf_params (((byte *) kdf_params) + 2, r_params);
}

// FIXME: too large?
#define KDF_PARAMS_MAX_LEN 512

/* Returns KEK parameters as an opaque MPI in R_PARAMS. */
gpg_error_t
pk_gost_default_params (const char *oidstr, unsigned int qbits,
                        gcry_mpi_t *r_params)
{
  gpg_error_t ret = GPG_ERR_NO_ERROR;
  unsigned char *kdf_params = xtrymalloc (KDF_PARAMS_MAX_LEN);
  size_t len = KDF_PARAMS_MAX_LEN - 2;

  kdf_params[1] = GOST_KDF_PARAMS_VERSION;

  /* Search for matching KEK parameter. Defaults to the strongest
     possible choices. */
  for (int i = 0; i < DIM (gost_kdf_params_table); i++)
    {
      if (0 == strncmp (oidstr, gost_kdf_params_table[i].oidpfx, strlen (gost_kdf_params_table[i].oidpfx))
          && gost_kdf_params_table[i].qbits >= qbits)
        {
          ret = pack_gost_kdf_params (&gost_kdf_params_table[i].params,
                                      &kdf_params[2],
                                      &len);
          break;
        }
    }

  if (ret != GPG_ERR_NO_ERROR)
    goto exit;

  if (len > 255)
    {
      ret = GPG_ERR_TOO_LARGE;
      goto exit;
    }

  kdf_params[0] = (byte) (len + 1);

  if (DBG_CRYPTO)
    log_printhex ("GOST KDF params are", kdf_params, len + 2);

  *r_params = gcry_mpi_set_opaque (*r_params, kdf_params, (len + 2) * 8);
  if (!*r_params)
    {
      ret = gpg_error_from_syserror ();
      goto exit;
    }

 exit:
  if (ret != GPG_ERR_NO_ERROR)
    {
      xfree (kdf_params);
    }

  return ret;
}

/* Generate an UKM value for a public ECDH (GOST) key in PKEY.
   On success the generated value is stored at R_UKM. On failure NULL
   is stored and an error code is returned. */
gpg_error_t
pk_gost_generate_ukm (gcry_mpi_t *pkey, gcry_mpi_t *r_ukm,
                      unsigned int *r_nbits)
{
  gost_kdf_params_t *params = NULL;
  gpg_error_t ret;

  *r_ukm = NULL;

  ret = pk_gost_get_kdf_params (pkey, &params);
  if (ret != GPG_ERR_NO_ERROR)
    return ret;

  int ukm_blen;
  switch (params->vko_algo)
    {
    case VKO_7836:
      ukm_blen = params->vko_params.vko_7836.ukm_len * 8;
      break;

    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
      goto exit;
    }

  if (!ukm_blen)
    {
      ret = GPG_ERR_TOO_SHORT;
      goto exit;
    }

  ret = gost_generate_ukm (ukm_blen, r_ukm);

 exit:
  free_gost_kdf_params (params);

  if (GPG_ERR_NO_ERROR == ret && r_nbits)
      *r_nbits = ukm_blen;

  return ret;
}

static const char*
digest_params_to_oid (digest_params_t digest_params)
{
  switch (digest_params)
    {
    case DIGEST_PARAMS_GOSTR3411_94_A:
      return "1.2.643.2.2.30.1";
    default:
      return NULL;
    }
}

static const char*
cipher_params_to_sbox (cipher_params_t cipher_params)
{
  switch (cipher_params)
    {
    case CIPHER_PARAMS_GOST28147_A:
      return "1.2.643.2.2.31.1";
    case CIPHER_PARAMS_GOST28147_B:
      return "1.2.643.2.2.31.2";
    case CIPHER_PARAMS_GOST28147_C:
      return "1.2.643.2.2.31.3";
    case CIPHER_PARAMS_GOST28147_D:
      return "1.2.643.2.2.31.4";
    case CIPHER_PARAMS_GOST28147_Z:
      return "1.2.643.7.1.2.5.1.1";
    default:
      return NULL;
    }
}

static const char*
mac_params_to_sbox (mac_params_t mac_params)
{
  switch (mac_params)
    {
    case MAC_PARAMS_GOST28147_A:
      return "1.2.643.2.2.31.1";
    case MAC_PARAMS_GOST28147_B:
      return "1.2.643.2.2.31.2";
    case MAC_PARAMS_GOST28147_C:
      return "1.2.643.2.2.31.3";
    case MAC_PARAMS_GOST28147_D:
      return "1.2.643.2.2.31.4";
    case MAC_PARAMS_GOST28147_Z:
      return "1.2.643.7.1.2.5.1.1";
    default:
      return NULL;
    }
}

static gpg_error_t
gost_kdf (const gost_kdf_params_t *params, gcry_mpi_t ukm,
          const unsigned char *kek_buf, size_t kek_len,
          gcry_mpi_t *r_kek)
{
  gpg_error_t ret = GPG_ERR_NO_ERROR;

  switch (params->kdf_algo)
    {
    case KDF_NULL:
      *r_kek = gcry_mpi_set_opaque_copy (*r_kek, kek_buf, 8 * kek_len);
      break;
    case GOST_KDF_CPDIVERS:
      ret = gost_cpdiversify_key (r_kek,
                                  map_cipher_openpgp_to_gcry (params->kdf_params.kdf_4357.kdf_cipher_algo),
                                  cipher_params_to_sbox (params->kdf_params.kdf_4357.kdf_cipher_params),
                                  kek_buf, kek_len, ukm);
      break;
    case GOST_KDF_TREE:
      /* TODO */
      ret = GPG_ERR_UNSUPPORTED_ALGORITHM;
      break;
    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
    }

  return ret;
}

static gpg_error_t
gost_vko_kdf (const gost_kdf_params_t *params, gcry_mpi_t shared,
              gcry_mpi_t ukm, gcry_mpi_t *r_kek)
{
  unsigned char *kek_buf = NULL;
  size_t kek_len;
  gpg_error_t ret = GPG_ERR_NO_ERROR;

  switch (params->vko_algo)
    {
    case VKO_7836:
      ret = gost_vko (shared,
                      map_md_openpgp_to_gcry (params->vko_params.vko_7836.vko_digest_algo),
                      digest_params_to_oid (params->vko_params.vko_7836.vko_digest_params),
                      &kek_buf, &kek_len);
      break;

    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
    }

  if (ret != GPG_ERR_NO_ERROR)
    goto exit;

  ret = gost_kdf (params, ukm, kek_buf, kek_len, r_kek);

 exit:
  xfree (kek_buf);

  return ret;
}

/* Encrypts a secret key DATA for the is the recepient public key PKEY using
   a key encryption key that is calculated from a SHARED point and a UKM value
   in accordance with the key KDF parameters.
   On success the result is stored at R_RESULT; on failure NULL is stored
   at R_RESULT and an error code returned.  */
gpg_error_t
pk_gost_encrypt_with_shared_point (gcry_mpi_t shared, gcry_mpi_t ukm,
                                   gcry_mpi_t data, gcry_mpi_t *pkey,
                                   gcry_mpi_t *r_result)
{
  gost_kdf_params_t *kdf_params = NULL;
  gcry_mpi_t kek = NULL;
  gcry_mpi_t encoded_key = NULL;
  unsigned char *encbuf = NULL;
  gpg_error_t ret = GPG_ERR_NO_ERROR;

  if (DBG_CRYPTO)
    {
      log_printmpi ("GOST unwrapped value:", data);
      log_printmpi ("GOST UKM:", ukm);
      log_printmpi ("GOST shared point:", shared);
    }

  *r_result = NULL;

  ret = pk_gost_get_kdf_params (pkey, &kdf_params);
  if (ret != GPG_ERR_NO_ERROR)
    return ret;

  ret = gost_vko_kdf (kdf_params, shared, ukm, &kek);
  if (ret != GPG_ERR_NO_ERROR)
    goto exit;

  if (DBG_CRYPTO)
    log_printmpi ("GOST KEK:", kek);

  switch (kdf_params->keywrap_algo)
    {
    case KEYWRAP_7836:
      ret = gost_keywrap (&encoded_key,
                          map_cipher_openpgp_to_gcry (kdf_params->keywrap_params.keywrap_7836.keywrap_cipher_algo),
                          cipher_params_to_sbox (kdf_params->keywrap_params.keywrap_7836.keywrap_cipher_params),
                          map_mac_openpgp_to_gcry (kdf_params->keywrap_params.keywrap_7836.keywrap_mac_algo),
                          mac_params_to_sbox (kdf_params->keywrap_params.keywrap_7836.keywrap_mac_params),
                          data, ukm, kek);
      break;
    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
    }
  if (ret != GPG_ERR_NO_ERROR)
    goto exit;

  if (DBG_CRYPTO)
    log_printmpi ("GOST wrapped value:", encoded_key);

  unsigned int enc_blen;
  unsigned char *encoded_key_buf = gcry_mpi_get_opaque (encoded_key, &enc_blen);
  size_t enclen = (enc_blen+7)/8;
  if (enclen > 255)
    {
      ret = GPG_ERR_TOO_LARGE;
      goto exit;
    }

  encbuf = xmalloc (enclen + 1);
  if (!encbuf)
    {
      ret = gpg_error_from_syserror ();
      goto exit;
    }

  encbuf[0] = (unsigned char) enclen;
  memcpy (encbuf + 1, encoded_key_buf, enclen);

  *r_result = gcry_mpi_set_opaque (*r_result, encbuf, (enclen + 1) * 8);

  if (DBG_CRYPTO)
    log_printmpi ("GOST encrypted value:", *r_result);

 exit:
  free_gost_kdf_params (kdf_params);
  gcry_mpi_release (kek);
  gcry_mpi_release (encoded_key);

  if (!*r_result)
    xfree (encbuf);

  return ret;
}

/* Decrypts a secret key DATA encrypted for the public key PKEY using
   a key encryption key that is calculated from a SHARED point and a UKM value
   in accordance with the key KDF parameters. On success the result is stored
   at R_RESULT; on failure NULL is stored at R_RESULT and an error code
   returned. */
gpg_error_t
pk_gost_decrypt_with_shared_point (gcry_mpi_t shared, gcry_mpi_t ukm,
                                   gcry_mpi_t data, gcry_mpi_t *pkey,
                                   gcry_mpi_t *r_result)
{
  gost_kdf_params_t *kdf_params = NULL;
  gcry_mpi_t kek = NULL;
  unsigned char *data_buf = NULL;
  gpg_error_t ret = GPG_ERR_NO_ERROR;

  if (DBG_CRYPTO)
    log_printmpi ("GOST encrypted value:", data);

  *r_result = NULL;

  unsigned int data_bits;
  data_buf = gcry_mpi_get_opaque (data, &data_bits);
  if (!data_buf)
    {
      ret = gpg_error_from_syserror ();
      goto exit;
    }

  byte data_len = (byte) ((data_bits + 7)/8);
  if (data_buf[0] != (data_len - 1))
    {
      ret = GPG_ERR_BAD_MPI;
      goto exit;
    }

  if (DBG_CRYPTO)
    {
      log_printhex ("GOST wrapped value:", data_buf + 1, data_len - 1);
      log_printmpi ("GOST UKM:", ukm);
    }

  ret = pk_gost_get_kdf_params (pkey, &kdf_params);
  if (ret != GPG_ERR_NO_ERROR)
    return ret;

  unsigned int shared_blen = gcry_mpi_get_nbits (shared);
  if (shared_blen < gcry_mpi_get_nbits (pkey[1]))
    {
      if (DBG_CRYPTO)
        log_debug ("GOST shared point: --\n");

      /* It seems that a KEK is directly passed here, possibly from
         a hardware token or card. */
      unsigned char shared_len = (shared_blen+7)/8;
      unsigned char *kek_buf = xtrymalloc_secure (shared_len);
      if (!kek_buf)
        {
          ret = gpg_error_from_syserror ();
          goto exit;
        }

      ret = gcry_mpi_print (GCRYMPI_FMT_USG, kek_buf, shared_len, NULL,
                            shared);
      if (ret != GPG_ERR_NO_ERROR)
        goto exit;

      ret = gost_kdf (kdf_params, ukm, kek_buf, shared_len, &kek);
      xfree (kek_buf);

      if (!kek)
        {
          ret = gpg_error_from_syserror ();
          goto exit;
        }
    }
  else
    {
      /* Normal shared point case. */
      if (DBG_CRYPTO)
        log_printmpi ("GOST shared point:", shared);

      ret = gost_vko_kdf (kdf_params, shared, ukm, &kek);
      if (ret != GPG_ERR_NO_ERROR)
        goto exit;
    }

  if (DBG_CRYPTO)
    log_printmpi ("GOST KEK:", kek);

  switch (kdf_params->keywrap_algo)
    {
    case KEYWRAP_7836:
      ret = gost_keyunwrap (r_result,
                          map_cipher_openpgp_to_gcry (kdf_params->keywrap_params.keywrap_7836.keywrap_cipher_algo),
                          cipher_params_to_sbox (kdf_params->keywrap_params.keywrap_7836.keywrap_cipher_params),
                          map_mac_openpgp_to_gcry (kdf_params->keywrap_params.keywrap_7836.keywrap_mac_algo),
                          mac_params_to_sbox (kdf_params->keywrap_params.keywrap_7836.keywrap_mac_params),
                            data_buf + 1, data_len - 1, ukm, kek);
      break;
    default:
      ret = GPG_ERR_UNKNOWN_ALGORITHM;
    }
  if (ret != GPG_ERR_NO_ERROR)
    goto exit;

  if (DBG_CRYPTO)
    log_printmpi ("GOST unwrapped value:", *r_result);

 exit:
  free_gost_kdf_params (kdf_params);
  gcry_mpi_release (kek);

  return ret;
}
