/* encrypt.c - Encrypt a message
 * Copyright (C) 2001, 2003, 2004, 2007, 2008,
 *               2010 Free Software Foundation, Inc.
 *
 * This file is part of GnuPG.
 *
 * GnuPG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnuPG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>

#include "gpgsm.h"
#include <gcrypt.h>
#include <ksba.h>

#include "keydb.h"
#include "../common/i18n.h"
#include "../common/compliance.h"
#include "../common/util.h"
#include "../common/gost-util.h"


struct dek_s {
  const char *algoid;
  const char *sbox_oid;
  int algo;
  int mode;
  gcry_cipher_hd_t chd;
  char key[32];
  int keylen;
  char iv[32];
  int ivlen;
};
typedef struct dek_s *DEK;


/* Callback parameters for the encryption.  */
struct encrypt_cb_parm_s
{
  estream_t fp;
  DEK dek;
  int eof_seen;
  int ready;
  int readerror;
  int bufsize;
  unsigned char *buffer;
  int buflen;
};





/* Initialize the data encryption key (session key). */
static int
init_dek (DEK dek)
{
  int rc=0, mode, i;

  dek->algo = gcry_cipher_map_name (dek->algoid);
  mode = gcry_cipher_mode_from_oid (dek->algoid);
  if (!dek->algo || !mode)
    {
      log_error ("unsupported algorithm '%s'\n", dek->algoid);
      return gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
    }

  /* Store the mode for use in encrypt_cb() to decide about
     the padding. */
  dek->mode = mode;

  /* Extra check for algorithms we consider to be too weak for
     encryption, although we support them for decryption.  Note that
     there is another check below discriminating on the key length. */
  switch (dek->algo)
    {
    case GCRY_CIPHER_DES:
    case GCRY_CIPHER_RFC2268_40:
      log_error ("cipher algorithm '%s' not allowed: too weak\n",
                 gnupg_cipher_algo_name (dek->algo));
      return gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
    default:
      break;
    }

  dek->keylen = gcry_cipher_get_algo_keylen (dek->algo);
  if (!dek->keylen || dek->keylen > sizeof (dek->key))
    return gpg_error (GPG_ERR_BUG);

  dek->ivlen = gcry_cipher_get_algo_blklen (dek->algo);
  if (!dek->ivlen || dek->ivlen > sizeof (dek->iv))
    return gpg_error (GPG_ERR_BUG);

  /* Make sure we don't use weak keys. */
  if (dek->keylen < 100/8)
    {
      log_error ("key length of '%s' too small\n", dek->algoid);
      return gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
    }

  rc = gcry_cipher_open (&dek->chd, dek->algo, mode, GCRY_CIPHER_SECURE);
  if (rc)
    {
      log_error ("failed to create cipher context: %s\n", gpg_strerror (rc));
      return rc;
    }

  for (i=0; i < 8; i++)
    {
      gcry_randomize (dek->key, dek->keylen, GCRY_STRONG_RANDOM );
      rc = gcry_cipher_setkey (dek->chd, dek->key, dek->keylen);
      if (gpg_err_code (rc) != GPG_ERR_WEAK_KEY)
        break;
      log_info(_("weak key created - retrying\n") );
    }
  if (rc)
    {
      log_error ("failed to set the key: %s\n", gpg_strerror (rc));
      goto leave;
    }

  /* Set the S-box if configured. */
  if (dek->sbox_oid)
    rc = gcry_cipher_ctl (dek->chd, GCRYCTL_SET_SBOX,
                          /* It's known that SET_SBOX doesn't modify
                             its args (FIXME). */
                          (char *) dek->sbox_oid,
                          strlen (dek->sbox_oid));

  if (rc)
	{
	  log_error("s-box setup failed: %s\n", gpg_strerror (rc));
	  goto leave;
	}

  gcry_create_nonce (dek->iv, dek->ivlen);
  rc = gcry_cipher_setiv (dek->chd, dek->iv, dek->ivlen);
  if (rc)
    {
      log_error ("failed to set the IV: %s\n", gpg_strerror (rc));
      goto leave;
    }

 leave:
  if (rc && dek->chd)
	{
	  gcry_cipher_close (dek->chd);
	  dek->chd = NULL;
	}

  return rc;
}


static int
encode_session_key (DEK dek, gcry_sexp_t * r_data)
{
  gcry_sexp_t data;
  int rc;

  switch (dek->algo)
	{
	case GCRY_CIPHER_GOST28147:
	  {
        rc = GPG_ERR_BUG;
	  }
	  break;
	default:
	  {
		char *p = xtrymalloc (64 + 2 * dek->keylen);
		if (!p)
		  return gpg_error_from_syserror ();
		strcpy (p, "(data\n (flags pkcs1)\n (value #");
		bin2hex (dek->key, dek->keylen, p + strlen (p));
		strcat (p, "#))\n");
		rc = gcry_sexp_sscan (&data, NULL, p, strlen (p));
		xfree (p);
	  }
	  break;
	}

  if (!rc)
	*r_data = data;
  return rc;
}


/* Encrypt the DEK under the key contained in CERT and return it as a
   canonical S-Exp in encval. */
static int
encrypt_dek (const DEK dek, ksba_cert_t cert, unsigned char **encval)
{
  gcry_sexp_t s_ciph = NULL, s_data = NULL, s_pkey = NULL;
  char *algo_oidstr = NULL;
  char *curve_oidstr = NULL;
  char *digest_oidstr = NULL;
  int rc;
  ksba_sexp_t buf;
  size_t len;

  *encval = NULL;

  /* get the key from the cert */
  buf = ksba_cert_get_public_key (cert);
  if (!buf)
    {
      log_error ("no public key for recipient\n");
      return gpg_error (GPG_ERR_NO_PUBKEY);
    }
  len = gcry_sexp_canon_len (buf, 0, NULL, NULL);
  if (!len)
    {
      log_error ("libksba did not return a proper S-Exp\n");
      return gpg_error (GPG_ERR_BUG);
    }
  rc = gcry_sexp_sscan (&s_pkey, NULL, (char*)buf, len);
  xfree (buf); buf = NULL;
  if (rc)
    {
      log_error ("gcry_sexp_scan failed: %s\n", gpg_strerror (rc));
      return rc;
    }

  unsigned int nbits;
  int pk_algo = gpgsm_get_key_algo_info (cert, &nbits, &algo_oidstr,
                                         &curve_oidstr, &digest_oidstr);

  /* Put the encoded cleartext into a simple list. */
  s_data = NULL; /* (avoid compiler warning) */

  switch (pk_algo)
    {
    case PUBKEY_ALGO_ECDH:
      {
        gcry_mpi_t ukm = NULL;
        unsigned int ukm_blen = 64;
        gcry_mpi_t privk = gcry_mpi_snew (nbits);
        if (!privk)
          {
            rc = gpg_error_from_syserror ();
            goto exit_ecdh;
          }
        gcry_mpi_randomize (privk, nbits-1, GCRY_STRONG_RANDOM);

        if (openpgp_oidstr_is_gost (algo_oidstr))
          {
            rc = gost_generate_ukm (ukm_blen, &ukm);
            if (rc) goto exit_ecdh;

            if (DBG_CRYPTO)
              log_printmpi ("UKM: ", ukm);

            gcry_mpi_lshift (privk, privk, ukm_blen);
            gcry_mpi_add (privk, privk, ukm);
            rc = gcry_sexp_build (&s_data, NULL, "(data(flags gost)(value %m)(salt-length %u))", privk, ukm_blen);
          }
        else
          rc = gcry_sexp_build (&s_data, NULL, "%m", privk);

        if (rc) goto exit_ecdh;

        rc = gcry_pk_encrypt (&s_ciph, s_data, s_pkey);
        if (rc) goto exit_ecdh;

        if (openpgp_oidstr_is_gost (algo_oidstr))
          {
            gcry_mpi_t key = NULL;
            gcry_mpi_t kek = NULL;
            gcry_mpi_t shared = get_mpi_from_sexp (s_ciph, "s", GCRYMPI_FMT_USG);
            gcry_mpi_t public = get_mpi_from_sexp (s_ciph, "e", GCRYMPI_FMT_USG);
            if (!shared || !public)
              {
                rc = GPG_ERR_BAD_DATA;
                goto exit_gost;
              }

            if (DBG_CRYPTO)
              log_printmpi ("Shared point: ", shared);

            rc = gcry_mpi_scan (&key, GCRYMPI_FMT_USG, dek->key,
                                dek->keylen, NULL);
            if (rc) goto exit_gost;

            enum gcry_md_algos digest_algo;
            enum gcry_cipher_algos cipher_algo;
            enum gcry_mac_algos mac_algo;
            const char *sbox = NULL;
            if (0 == strcmp (algo_oidstr, "1.2.643.2.2.19"))
              {
                digest_algo = GCRY_MD_GOSTR3411_CP;
                cipher_algo = GCRY_CIPHER_GOST28147;
                mac_algo = GCRY_MAC_GOST28147_IMIT;
                sbox = "1.2.643.2.2.31.1";
              }
            else if (0 == strcmp (algo_oidstr, "1.2.643.7.1.1.1.1") ||
                     0 == strcmp (algo_oidstr, "1.2.643.7.1.1.1.2"))
              {
                digest_algo = GCRY_MD_STRIBOG256;
                cipher_algo = GCRY_CIPHER_GOST28147;
                mac_algo = GCRY_MAC_GOST28147_IMIT;
                sbox = "1.2.643.7.1.2.5.1.1";
              }
            else
              {
                rc = GPG_ERR_UNSUPPORTED_ALGORITHM;
                goto exit_gost;
              }

            if (DBG_CRYPTO)
              {
                log_debug ("KEK digest: %s\n", gcry_md_algo_name (digest_algo));
                log_debug ("KEK S-box: %s\n", sbox);
              }

            unsigned char *kek_buf = NULL;
            size_t kek_len;
            rc = gost_vko (shared, digest_algo, NULL, &kek_buf, &kek_len);
            if (rc) goto exit_gost;

            if (DBG_CRYPTO)
                log_printhex ("KEK (raw): ", kek_buf, kek_len);

            /* Opaquefy the UKM value in order to properly store a value
               with leading zeros. */
            size_t ukm_len = (ukm_blen + 7)/8;
            unsigned char *ukm_buf = xtrymalloc (ukm_len);
            if (!ukm_buf)
              {
                rc = gpg_error_from_syserror ();
                goto exit_gost;
              }
            memset (ukm_buf, 0, ukm_len);
            size_t ukm_wrt;
            rc = gcry_mpi_print (GCRYMPI_FMT_USG, ukm_buf, ukm_len, &ukm_wrt, ukm);
            if (ukm_wrt < ukm_len)
              {
                memmove (ukm_buf + (ukm_len - ukm_wrt), ukm_buf, ukm_wrt);
                memset (ukm_buf, 0, ukm_len - ukm_wrt);
              }
            ukm = gcry_mpi_set_opaque (ukm, ukm_buf, ukm_blen);

            rc = gost_cpdiversify_key (&kek, cipher_algo, sbox,
                                       kek_buf, kek_len, ukm);
            if (rc) goto exit_gost;

            if (DBG_CRYPTO)
              log_printmpi ("KEK: ", kek);

            gcry_mpi_t wrapped = NULL;
            rc = gost_keywrap (&wrapped, cipher_algo, sbox, mac_algo, sbox,
                               key, ukm, kek);
            if (rc) goto exit_gost;

            if (DBG_CRYPTO)
              log_printmpi ("Wrapped DEK: ", wrapped);

            gcry_sexp_release (s_ciph);
            rc = gcry_sexp_build (&s_ciph, NULL, "(enc-val(gost(q %m)(ukm %m)(s %m)(algo %s)(curve %s)(digest %s)(sbox %s)))", public, ukm, wrapped, algo_oidstr, curve_oidstr, digest_oidstr, sbox);

          exit_gost:
            gcry_mpi_release (shared);
            gcry_mpi_release (public);
            gcry_mpi_release (kek);
            gcry_mpi_release (key);
          }
        else
          {
            /* TODO: ECDH encryption */
            rc = GPG_ERR_UNSUPPORTED_ALGORITHM;
          }

      exit_ecdh:
        gcry_mpi_release (ukm);
        gcry_mpi_release (privk);
      }
      break;
    default:
      rc = encode_session_key (dek, &s_data);
      if (!rc)
        rc = gcry_pk_encrypt (&s_ciph, s_data, s_pkey);
    }

  if (!rc)
    rc = make_canon_sexp (s_ciph, encval, NULL);

  if (rc)
    log_error ("encode_session_key failed: %s\n", gpg_strerror (rc));

  gcry_sexp_release (s_ciph);
  gcry_sexp_release (s_data);
  gcry_sexp_release (s_pkey);

  xfree (algo_oidstr);
  xfree (curve_oidstr);
  xfree (digest_oidstr);

  return rc;
}



/* do the actual encryption */
static int
encrypt_cb (void *cb_value, char *buffer, size_t count, size_t *nread)
{
  struct encrypt_cb_parm_s *parm = cb_value;
  int blklen = parm->dek->ivlen;
  unsigned char *p;
  size_t n;

  *nread = 0;
  if (!buffer)
    return -1; /* not supported */

  if (parm->ready)
    return -1;

  if (count < blklen)
    BUG ();

  if (!parm->eof_seen)
    { /* fillup the buffer */
      p = parm->buffer;
      for (n=parm->buflen; n < parm->bufsize; n++)
        {
          int c = es_getc (parm->fp);
          if (c == EOF)
            {
              if (es_ferror (parm->fp))
                {
                  parm->readerror = errno;
                  return -1;
                }
              parm->eof_seen = 1;
              break;
            }
          p[n] = c;
        }
      parm->buflen = n;
    }

  n = parm->buflen < count? parm->buflen : count;
  n = n/blklen * blklen;
  if (n)
    { /* encrypt the stuff */
      gcry_cipher_encrypt (parm->dek->chd, buffer, n, parm->buffer, n);
      *nread = n;
      /* Who cares about cycles, take the easy way and shift the buffer */
      parm->buflen -= n;
      if (parm->buflen)
        memmove (parm->buffer, parm->buffer+n, parm->buflen);
      else if (parm->eof_seen)
        {
          switch (parm->dek->mode)
            { /* Set the 'ready' flag because no padding is required
                 for these modes. */
            case GCRY_CIPHER_MODE_CFB:
            case GCRY_CIPHER_MODE_OFB:
            case GCRY_CIPHER_MODE_CTR:
              parm->ready = 1;
              break;
            default:
              break;
            }
        }
    }
  else if (parm->eof_seen)
    { /* no complete block but eof: add padding */
      /* fixme: we should try to do this also in the above code path */
      n = parm->buflen;
      switch (parm->dek->mode)
        { /* No padding is required for these modes */
        case GCRY_CIPHER_MODE_CFB:
        case GCRY_CIPHER_MODE_OFB:
        case GCRY_CIPHER_MODE_CTR:
          break;
        default:
          {
            int i, npad = blklen - (parm->buflen % blklen);
            p = parm->buffer;
            for (i=0; n < parm->bufsize && i < npad; n++, i++)
              p[n] = npad;
          }
          break;
        }
      gcry_cipher_encrypt (parm->dek->chd, buffer, n, parm->buffer, n);
      *nread = n;
      parm->ready = 1;
    }

  return 0;
}




/* Perform an encrypt operation.

   Encrypt the data received on DATA-FD and write it to OUT_FP.  The
   recipients are take from the certificate given in recplist; if this
   is NULL it will be encrypted for a default recipient */
int
gpgsm_encrypt (ctrl_t ctrl, certlist_t recplist, int data_fd, estream_t out_fp)
{
  int rc = 0;
  gnupg_ksba_io_t b64writer = NULL;
  gpg_error_t err;
  ksba_writer_t writer;
  ksba_reader_t reader = NULL;
  ksba_cms_t cms = NULL;
  ksba_stop_reason_t stopreason;
  KEYDB_HANDLE kh = NULL;
  struct encrypt_cb_parm_s encparm;
  DEK dek = NULL;
  int recpno;
  estream_t data_fp = NULL;
  certlist_t cl;
  int count;
  int compliant;

  memset (&encparm, 0, sizeof encparm);

  audit_set_type (ctrl->audit, AUDIT_TYPE_ENCRYPT);

  /* Check that the certificate list is not empty and that at least
     one certificate is not flagged as encrypt_to; i.e. is a real
     recipient. */
  for (cl = recplist; cl; cl = cl->next)
    if (!cl->is_encrypt_to)
      break;
  if (!cl)
    {
      log_error(_("no valid recipients given\n"));
      gpgsm_status (ctrl, STATUS_NO_RECP, "0");
      audit_log_i (ctrl->audit, AUDIT_GOT_RECIPIENTS, 0);
      rc = gpg_error (GPG_ERR_NO_PUBKEY);
      goto leave;
    }

  for (count = 0, cl = recplist; cl; cl = cl->next)
    count++;
  audit_log_i (ctrl->audit, AUDIT_GOT_RECIPIENTS, count);

  kh = keydb_new ();
  if (!kh)
    {
      log_error (_("failed to allocate keyDB handle\n"));
      rc = gpg_error (GPG_ERR_GENERAL);
      goto leave;
    }

  /* Fixme:  We should use the unlocked version of the es functions.  */
  data_fp = es_fdopen_nc (data_fd, "rb");
  if (!data_fp)
    {
      rc = gpg_error_from_syserror ();
      log_error ("fdopen() failed: %s\n", strerror (errno));
      goto leave;
    }

  err = ksba_reader_new (&reader);
  if (err)
      rc = err;
  if (!rc)
    rc = ksba_reader_set_cb (reader, encrypt_cb, &encparm);
  if (rc)
      goto leave;

  encparm.fp = data_fp;

  ctrl->pem_name = "ENCRYPTED MESSAGE";
  rc = gnupg_ksba_create_writer
    (&b64writer, ((ctrl->create_pem? GNUPG_KSBA_IO_PEM : 0)
                  | (ctrl->create_base64? GNUPG_KSBA_IO_BASE64 : 0)),
     ctrl->pem_name, out_fp, &writer);
  if (rc)
    {
      log_error ("can't create writer: %s\n", gpg_strerror (rc));
      goto leave;
    }

  err = ksba_cms_new (&cms);
  if (err)
    {
      rc = err;
      goto leave;
    }

  err = ksba_cms_set_reader_writer (cms, reader, writer);
  if (err)
    {
      log_debug ("ksba_cms_set_reader_writer failed: %s\n",
                 gpg_strerror (err));
      rc = err;
      goto leave;
    }

  audit_log (ctrl->audit, AUDIT_GOT_DATA);

  /* We are going to create enveloped data with uninterpreted data as
     inner content */
  err = ksba_cms_set_content_type (cms, 0, KSBA_CT_ENVELOPED_DATA);
  if (!err)
    err = ksba_cms_set_content_type (cms, 1, KSBA_CT_DATA);
  if (err)
    {
      log_debug ("ksba_cms_set_content_type failed: %s\n",
                 gpg_strerror (err));
      rc = err;
      goto leave;
    }

  /* Check compliance.  */
  if (!gnupg_cipher_is_allowed
      (opt.compliance, 1, gcry_cipher_map_name (opt.def_cipher_algoid),
       gcry_cipher_mode_from_oid (opt.def_cipher_algoid)))
    {
      log_error (_("cipher algorithm '%s' may not be used in %s mode\n"),
		 opt.def_cipher_algoid,
		 gnupg_compliance_option_string (opt.compliance));
      rc = gpg_error (GPG_ERR_CIPHER_ALGO);
      goto leave;
    }

  if (!gnupg_rng_is_compliant (opt.compliance))
    {
      rc = gpg_error (GPG_ERR_FORBIDDEN);
      log_error (_("%s is not compliant with %s mode\n"),
                 "RNG",
                 gnupg_compliance_option_string (opt.compliance));
      gpgsm_status_with_error (ctrl, STATUS_ERROR,
                               "random-compliance", rc);
      goto leave;
    }

  dek = xtrycalloc_secure (1, sizeof *dek);
  if (!dek)
	{
	  rc = out_of_core ();
	  goto leave;
	}

  dek->algoid = NULL;
  for (recpno = 0, cl = recplist; cl; recpno++, cl = cl->next)
    {
      char *oidstr = NULL;
      int pk_algo = gpgsm_get_key_algo_info (cl->cert, NULL, &oidstr, NULL,
                                             NULL);
	  switch (pk_algo)
		{
		case PUBKEY_ALGO_ECDH:
		  if (openpgp_oidstr_is_gost (oidstr))
			{
              if (!dek->algoid)
                dek->algoid = "1.2.643.2.2.21";
              else if (0 == strcmp (dek->algoid, "1.2.643.2.2.21"))
                break;
              else
                {
                  rc = gpg_error (GPG_ERR_CIPHER_ALGO);
                  goto dek_leave;
                }

              if (!dek->sbox_oid)
                {
                  if (0 == strcmp (oidstr, "1.2.643.7.1.1.1.1") ||
                      0 == strcmp (oidstr, "1.2.643.7.1.1.1.2"))
                    {
                      /* GOST 2012 */
                      dek->sbox_oid = "1.2.643.7.1.2.5.1.1";
                    }
                  else if (0 == strcmp (oidstr, "1.2.643.2.2.19"))
                    {
                      /* GOST 2001 */
                      dek->sbox_oid = "1.2.643.2.2.31.1";
                    }
                  else
                    {
                      rc = gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
                      goto dek_leave;
                    }
                }
			}
		  break;
		}

    dek_leave:
      xfree (oidstr);
      if (rc) goto leave;
	}

  if (!dek->algoid)
    dek->algoid = opt.def_cipher_algoid;

  /* Create a session key */
  rc = init_dek (dek);
  if (rc)
    {
      log_error ("failed to create the session key: %s\n",
                 gpg_strerror (rc));
      goto leave;
    }

  err = ksba_cms_set_content_enc_algo (cms,
                                       /* Use S-box OID to identify both
                                          the algorithm and the S-box
                                          (FIXME). */
                                       dek->sbox_oid ?
                                         dek->sbox_oid :
                                         dek->algoid,
                                       dek->iv, dek->ivlen);
  if (err)
    {
      log_error ("ksba_cms_set_content_enc_algo failed: %s\n",
                 gpg_strerror (err));
      rc = err;
      goto leave;
    }

  encparm.dek = dek;
  /* Use a ~8k (AES) or ~4k (3DES) buffer */
  encparm.bufsize = 500 * dek->ivlen;
  encparm.buffer = xtrymalloc (encparm.bufsize);
  if (!encparm.buffer)
    {
      rc = out_of_core ();
      goto leave;
    }

  audit_log_s (ctrl->audit, AUDIT_SESSION_KEY, dek->algoid);

  compliant = gnupg_cipher_is_compliant (CO_DE_VS, dek->algo,
                                         GCRY_CIPHER_MODE_CBC);

  /* Gather certificates of recipients, encrypt the session key for
     each and store them in the CMS object */
  for (recpno = 0, cl = recplist; cl; recpno++, cl = cl->next)
    {
      unsigned char *encval;
      unsigned int nbits;
      char *curve_oidstr = NULL;
      int pk_algo;

      /* Check compliance.  */
      pk_algo = gpgsm_get_key_algo_info (cl->cert, &nbits, NULL, &curve_oidstr, NULL);
      if (!gnupg_pk_is_compliant (opt.compliance, pk_algo, NULL, nbits, curve_oidstr))
        {
          char  kidstr[10+1];

          snprintf (kidstr, sizeof kidstr, "0x%08lX",
                    gpgsm_get_short_fingerprint (cl->cert, NULL));
          log_info (_("WARNING: key %s is not suitable for encryption"
                      " in %s mode\n"),
                    kidstr,
                    gnupg_compliance_option_string (opt.compliance));
        }

      if (compliant
          && !gnupg_pk_is_compliant (CO_DE_VS, pk_algo, NULL, nbits, curve_oidstr))
        compliant = 0;

      rc = encrypt_dek (dek, cl->cert, &encval);
      if (rc)
        {
          audit_log_cert (ctrl->audit, AUDIT_ENCRYPTED_TO, cl->cert, rc);
          log_error ("encryption failed for recipient no. %d: %s\n",
                     recpno, gpg_strerror (rc));
          goto setval_leave;
        }

      err = ksba_cms_add_recipient (cms, cl->cert);
      if (err)
        {
          audit_log_cert (ctrl->audit, AUDIT_ENCRYPTED_TO, cl->cert, err);
          log_error ("ksba_cms_add_recipient failed: %s\n",
                     gpg_strerror (err));
          rc = err;
          xfree (encval);
          goto setval_leave;
        }

      err = ksba_cms_set_enc_val (cms, recpno, encval);
      xfree (encval);
      audit_log_cert (ctrl->audit, AUDIT_ENCRYPTED_TO, cl->cert, err);
      if (err)
        {
          log_error ("ksba_cms_set_enc_val failed: %s\n",
                     gpg_strerror (err));
          rc = err;
          goto setval_leave;
        }

    setval_leave:
      xfree (curve_oidstr);
      if (rc) goto leave;
    }

  if (compliant)
    gpgsm_status (ctrl, STATUS_ENCRYPTION_COMPLIANCE_MODE,
                  gnupg_status_compliance_flag (CO_DE_VS));

  /* Main control loop for encryption. */
  recpno = 0;
  do
    {
      err = ksba_cms_build (cms, &stopreason);
      if (err)
        {
          log_debug ("ksba_cms_build failed: %s\n", gpg_strerror (err));
          rc = err;
          goto leave;
        }
    }
  while (stopreason != KSBA_SR_READY);

  if (encparm.readerror)
    {
      log_error ("error reading input: %s\n", strerror (encparm.readerror));
      rc = gpg_error (gpg_err_code_from_errno (encparm.readerror));
      goto leave;
    }


  rc = gnupg_ksba_finish_writer (b64writer);
  if (rc)
    {
      log_error ("write failed: %s\n", gpg_strerror (rc));
      goto leave;
    }
  audit_log (ctrl->audit, AUDIT_ENCRYPTION_DONE);
  log_info ("encrypted data created\n");

 leave:
  ksba_cms_release (cms);
  gnupg_ksba_destroy_writer (b64writer);
  ksba_reader_release (reader);
  keydb_release (kh);
  xfree (dek);
  es_fclose (data_fp);
  xfree (encparm.buffer);
  return rc;
}
